import json
from collections import defaultdict
from logging import Formatter, LogRecord
from traceback import format_exception
from typing import Iterable

__all__ = ["Office365CardFormatter", "TeamsCardsFormatter"]


class TeamsCardsFormatter(Formatter):
    """
    This class is the base class for cards formatters.
    https://docs.microsoft.com/en-us/microsoftteams/platform/task-modules-and-cards/what-are-cards
    """

    def format(self, record: LogRecord) -> str:
        raise NotImplementedError()


class Office365CardFormatter(TeamsCardsFormatter):
    """
    This formatter formats logs records as a simple office 365 connector card.
    The connector card documentation is displayed in the link below:
    https://docs.microsoft.com/en-us/microsoftteams/platform/task-modules-and-cards/cards/cards-reference#office-365-connector-card
    In addition to the message, each log record attribute (levelname, lineno...etc)
    can be displayed as facts.
    """

    _facts = {"name", "levelname", "levelno", "lineno"}
    _color_map = defaultdict(
        lambda: "#008000",
        {
            "DEBUG": "#0000FF",  # blue
            "INFO": "#008000",  # green
            "WARNING": "#FFA500",  # orange
            "ERROR": "#FF0000",  # red
            "CRITICAL": "#8B0000",  # darkred
        },
    )

    def __init__(self, facts: Iterable[str]):
        """
        :param facts:  LogRecord attributes to be displayed as facts in
                       the message's card.
        """
        self.facts = self._facts.intersection(set(facts))
        super().__init__()

    def format(self, record: LogRecord) -> str:
        message = record.getMessage()
        if record.exc_info:
            etype, value, tb = record.exc_info
            message += "\n" * 2
            message += "<code>"
            message += "".join(format_exception(etype, value, tb))
            message += "</code>"
        return json.dumps(
            {
                "@context": "https://schema.org/extensions",
                "@type": "MessageCard",
                "title": f"{record.levelname.title()} in {record.module}",
                "summary": f"{record.getMessage()}",
                "sections": [{"facts": self._build_facts_list(record)}],
                # Fallback to INFO color if needed
                "themeColor": self._color_map[record.levelname],
                "text": message,
            }
        )

    def _build_facts_list(self, record: LogRecord):
        return [{"name": fact, "value": getattr(record, fact)} for fact in self.facts]



import logging


class LoggingMixin:
    """ Convenience super-class to have a logger configured with the class name
    """

    def __init__(self, context=None):
        self._set_context(context)

    @property
    def log(self) -> logging.Logger:
        """ A logger object.
        Returns:
            Logger
        """
        self._log = logging.getLogger(
            self.__class__.__module__ + "." + self.__class__.__name__
        )
        # self.log.setup_logging(log_level=10)
        return self._log

    def _set_context(self, context):
        if context is not None:
            self.set_context(self.log, context)

    @staticmethod
    def set_context(logger, value):
        """ Walks the tree of loggers and tries to set the context for each handler.
        
        Args:
            logger: logger to traverse
            value: value to set
        """
        _logger = logger
        while _logger:
            for handler in _logger.handlers:
                try:
                    handler.set_context(value)
                except AttributeError:
                    # Not all handlers need to have context passed in so we ignore
                    # the error when handlers do not have set_context defined.
                    pass
            if _logger.propagate is True:
                _logger = _logger.parent
            else:
                _logger = None




import json
import os
import queue
from logging import NOTSET, Handler, LogRecord
from logging.handlers import QueueHandler, QueueListener

import requests

from gytrash.formatters import TeamsCardsFormatter

__all__ = [
    "TeamsHandler",
    "TeamsQueueHandler",
]


class TeamsHandler(Handler):
    """
    Logging handler for Microsoft Teams webhook integration.

    This handler blocks operation. Use non-blocking NBTeamsHandler for less
    impact on your application performance.
    """

    def __init__(self, url, level=NOTSET):
        """
        :param url: Microsoft Teams incoming webhook url.
        :param level: Logging level (INFO, DEBUG, ERROR...etc)
        """
        super().__init__(level=level)
        if url:
            self.url = url
        else:
            self.url = os.environ["TEAMS_CHANNEL_URL"]

    def format(self, record: LogRecord) -> str:
        if not isinstance(self.formatter, TeamsCardsFormatter):
            return json.dumps({"text": super().format(record)})
        else:
            return self.formatter.format(record)

    def emit(self, record: LogRecord):
        try:
            data = self.format(record)
            requests.post(
                url=self.url, headers={"Content-Type": "application/json"}, data=data
            )
        except Exception:
            self.handleError(record)


class TeamsQueueHandler(QueueHandler):
    """
    Non-blocking logging handler for Microsoft Teams webhook integration.

    This handler reduces impact on your application compared to the (blocking)
    TeamsHandler.

    Queue-based implementation from
    https://docs.python.org/3/howto/logging-cookbook.html#dealing-with-handlers-that-block
    """

    def __init__(self, url, level=NOTSET):
        self._log_queue = queue.Queue(-1)
        super().__init__(self._log_queue)

        self._teams_handler = TeamsHandler(url, level)
        teams_log_listener = QueueListener(self._log_queue, self._teams_handler)
        teams_log_listener.start()

    def setFormatter(self, fmt):
        self._teams_handler.setFormatter(fmt)
        super().setFormatter(fmt)



import logging
from re import S


class SlackFormatter(logging.Formatter):
    """Formatter constructs and formats json payload for the slack log handler.

    Inherits:
        logging.Formatter: Base log formatting class
    """

    ERROR_TYPES_EMOJI = {
        "CRITICAL": ":interrobang:",
        "ERROR": ":red_circle:",
        "WARNING": ":large_orange_diamond:",
        "INFO": ":large_blue_circle:",
        "DEBUG": ":white_circle:",
    }

    DIVIDER_BLOCK = {"type": "divider"}

    def format(self, record: "logging.LogRecord"):
        """Formats a message. Required method overwrite of base class.

        Args:
            record (logging.LogRecord): A log delivered from the handler.

        Returns:
            dict: dictionary representation of a slack payload.
        """
        record.asctime = self.formatTime(record)
        record.message = record.getMessage()

        ret = {
            "ts": "",
            "icon_emoji": self.ERROR_TYPES_EMOJI[record.levelname],
            "blocks": [*self._create_log_block(record), self.DIVIDER_BLOCK,],
        }
        return ret

    def _create_log_block(self, record: "logging.LogRecord"):
        """Creates a slack message "block".
        
        [Slack Block Kit](https://api.slack.com/block-kit)

        Args:
            record (logging.LogRecord): Logging record passed in for formatting.

        Returns:
            list(dict): Returns a list of dicts through a constructor function, _get_log_block.
        """
        emoji = f"{self.ERROR_TYPES_EMOJI[record.levelname]}"
        text = f"```{record.message}```"
        information = f"{record.asctime}-{record.name}[{record.process}]-{record.module}-{record.levelname}"
        return self._get_log_block(emoji, text, information)

    @staticmethod
    def _get_log_block(emoji: str, text: str, information: str):
        """Creates a slack payload block object.

        Args:
            emoji (str): An emoji colon bound string for use in the message block.
            text (str): The text of the message to include in the message block
            information (str): The metadata about the logging message.

        Returns:
            list(dict): List of dictionaries, one for each section of the message block.
        """
        return [
            {"type": "section", "text": {"type": "mrkdwn", "text": emoji}},
            {"type": "context", "elements": [{"type": "mrkdwn", "text": information}],},
            {"type": "section", "text": {"type": "mrkdwn", "text": text}},
        ]



import logging

import coloredlogs

from gytrash.__about__ import *  # noqa: F403,F401
from gytrash.filters import MessengerLogFilter
from gytrash.formatters import Office365CardFormatter, SlackFormatter
from gytrash.handlers import SlackHandler, TeamsHandler, TeamsQueueHandler

log = logging.getLogger("gytrash")


def setup_logging(
    log,
    *,
    log_level: int = 10,
    log_from_botocore: bool = False,
    log_to_slack: bool = False,
    log_to_teams: bool = False,
    slack_log_channel: str = None,
    slack_log_level: int = 20,
    slack_bot_token: str = None,
    teams_url: str = None,
    teams_log_level: int = 20,
    teams_nonblocking: bool = True,
    teams_card_formatter: bool = True,
) -> None:
    """Create the Logging handler for the CLI.
        This setups a log handler that support logging in color.

    Args:
        log: Root logging object.
        log_level: int - (keyword) console streamhandler log level
        log_from_botocore: bool - (keyword) Add botocore Logger if using boto
        log_to_slack: bool - (keyword) Add custom streamhandler to log to slack channel
        slack_log_channel: str - (keyword) Name of the slack channel to send logs
        slack_log_level: int - (keyword) slack streamhandler log level
        slack_bot_token: str - (keyword) Bot token to connect to a slack app.
        teams_url: str - (keyword) URL to the Teams incoming webhook for a
                         channel.
        teams_log_level: int - (keyword) Logging level for the teams logger.
        teams_nonblocking: bool - (keyword) Boolean to turn on non-blocking
                                  queue for log delivery.
        teams_card_formatter: bool - (keyword) Boolean for using the Office 365
                                     card format when delivering logs
    Returns:
        None
    """

    log_format = "%(asctime)s %(name)s:%(module)s:%(lineno)d[%(process)d]:: %(levelname)s %(message)s"  # noqa: E501

    log.setLevel(log_level)

    # generic_formatter = logging.Formatter(log_format)

    coloredlogs.install(level=log_level, logger=log, fmt=log_format)
    log.debug(f"Gytrash log level: {log.getEffectiveLevel()}")

    if log_from_botocore:
        log.debug("Tapping Botocore logger.")
        coloredlogs.install(
            level=log_level, logger=logging.getLogger("botocore"), fmt=log_format
        )

    if log_to_slack is True:
        sh = SlackHandler(slack_log_channel, slack_bot_token)
        log.addHandler(sh)
        sf = SlackFormatter(log_format)
        sh.setFormatter(sf)
        sfilt = MessengerLogFilter()
        sh.addFilter(sfilt)
        sh.setLevel(slack_log_level)

    if log_to_teams is True:
        if teams_nonblocking is True:
            th = TeamsQueueHandler(url=teams_url, level=teams_log_level)
        else:
            th = TeamsHandler(url=teams_url, level=teams_log_level)

        if teams_card_formatter is True:
            cf = Office365CardFormatter(facts=["name", "levelname", "lineno"])
            th.setFormatter(cf)

        log.addHandler(th)
        tfilt = MessengerLogFilter()
        th.addFilter(tfilt)



from logging import (
    StreamHandler,
    CRITICAL,
    ERROR,
    WARNING,
    INFO,
    FATAL,
    DEBUG,
    NOTSET,
    Formatter,
)

import logging
import os
import traceback

from slack_sdk.web import WebClient

ERROR_COLOR = "danger"  # color name is built in to Slack API
WARNING_COLOR = "warning"  # color name is built in to Slack API
INFO_COLOR = "#439FE0"

COLORS = {
    CRITICAL: ERROR_COLOR,
    FATAL: ERROR_COLOR,
    ERROR: ERROR_COLOR,
    WARNING: WARNING_COLOR,
    INFO: INFO_COLOR,
    DEBUG: INFO_COLOR,
    NOTSET: INFO_COLOR,
}


class SlackHandler(StreamHandler):
    """Slack log handler Class

    Inherits:
        logging.StreamHandler: Base log StreamHandler class.
    """

    def __init__(
        self, channel: str, slack_bot_token: str = None, username: str = "Gytrash"
    ):
        """Initialize the stream handler with some specifics for slack.

        Args:
            channel (str): Slack channel to publish logs
            slack_bot_token (str, optional): Slack bot token to use the slack published app. Defaults to None.
            username (str, optional): Username of the Slack bot. Defaults to "Gytrash".
        """
        StreamHandler.__init__(self)
        # Initialize a Web API client
        if slack_bot_token:
            self.slack_web_client = WebClient(token=slack_bot_token)
        else:
            self.slack_web_client = WebClient(token=os.environ["SLACK_BOT_TOKEN"])
        self.channel = channel
        self.username = username

    def _send_log(self, message: dict):
        """Posts the formatted message to slack via the web client.

        Args:
            message (dict): Slack message dictionary. Follows the blocks API.
        """
        self.slack_web_client.chat_postMessage(**message)

    def emit(self, message: "logging.LogRecord"):
        """Emits a message from the handler.

        Args:
            message (logging.LogRecord): Log record from the stream.
        """
        assert isinstance(message, logging.LogRecord)

        slack_message = self.format(message)

        # List of LogRecord attributes expected when reading the
        # documentation of the logging module:
        expected_attributes = (
            "args,asctime,created,exc_info,filename,funcName,levelname,"
            "levelno,lineno,module,msecs,message,msg,name,pathname,"
            "process,processName,relativeCreated,stack_info,thread,threadName"
        )
        for ea in expected_attributes.split(","):
            if not hasattr(message, ea):
                print("UNEXPECTED: LogRecord does not have the '{}' field!".format(ea))

        slack_message["channel"] = self.channel
        slack_message["username"] = self.username

        self._send_log(slack_message)



#!/usr/bin/env python

import configparser
import os
from codecs import open

from setuptools import find_packages, setup

# Import the package module code from local directory.
repo_path = os.path.abspath(os.path.dirname(__file__))

packages = find_packages(exclude=("examples",))

# Open requirements files to grab package dependencies
with open("requirements.txt") as f:
    install_requires = f.read().splitlines()

with open("requirements-test.txt") as f:
    tests_require = [line for line in f if "==" in line]

with open("requirements-dev.txt") as f:
    dev_requires = [line for line in f if "==" in line]

# Open README file to attach readme as package description.
with open("README.md", "r", "utf-8") as f:
    readme = f.read()


# with open(os.path.join(repo_path, "gytrash", "__about__.py"), "r", "utf-8") as f:
about = configparser.ConfigParser()
about.read(os.path.join(repo_path, "pyproject.toml"))

setup(
    name=about["default"]["__title__"].strip('"'),
    version=about["default"]["__version__"].strip('"'),
    description=about["default"]["__description__"].strip('"'),
    long_description=readme,
    long_description_content_type="text/markdown",
    author=about["default"]["__author__"].strip('"'),
    author_email=about["default"]["__author_email__"].strip('"'),
    url=about["default"]["__url__"].strip('"'),
    packages=packages,
    package_dir={"gytrash": "gytrash"},
    package_data={"": []},
    include_package_data=True,
    python_requires=">=3.7",
    install_requires=install_requires,
    setup_requires=install_requires,
    extras_require={"test": tests_require, "dev": dev_requires},
    license=about["default"]["__license__"].strip('"'),
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    tests_require=tests_require,
    project_urls={"Source": "https://github.com/trejas/gytrash"},
)



import logging


class MessengerLogFilter(logging.Filter):
    """
    Logging filter to decide when logging to a Messenger is requested, using
    the `extra` kwargs:
        `logger.info("...", extra={'notify_messenger': True})`

    Inherits:
        logging.Filter: Base log filtering class
    """

    def filter(self, record):
        return getattr(record, "notify_messenger", False)
